#!/usr/bin/perl
# ARCADEmini binary randomizer: ambinrnd
# version 1.1 date 2005-06-04
# part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
# written by 1stein <1stein@schuermans.info>
# Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html

use strict;

print <<EOF;
ARCADEmini binary randomizer: ambinrnd
version 1.1 date 2005-06-04
part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
written by 1stein <1stein\@schuermans.info>
Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html

EOF

if( @ARGV < 2 )
{
	print <<EOF;
syntax: ambinrnd <ambin-input-file 1> ... <ambin-input-file n> <ambin-output-file>

examples: ambinrnd in1.ambin in2.ambin out.ambin

EOF
	exit;
}

my $ident = "ARCADEmini 26x20/16 ambin 1.0   ";
my $i;
for( $i = 32; $i < 512; $i++ ) { $ident = $ident . pack( 'C', 0xFF ); }

# check input files and get information

my $file_cnt = @ARGV - 1;
my @files = ();
my $file_no;
for( $file_no = 0; $file_no < $file_cnt; $file_no++ )
{
  my $file = $ARGV[$file_no];
  if( open( InF, "< " . $file ) )
  {
    binmode InF;

    seek( InF, 0, 2 );
    my $file_size = tell( InF );
    my $file_sectors = ($file_size >> 9) - 2;

    seek( InF, 0, 0 );
    my $file_ident;
    read( InF, $file_ident, 512 );

    seek( InF, ($file_sectors << 9) + 512, 0 );
    my $file_end;
    read( InF, $file_end, 512 );
    my $file_end_1 = unpack( 'C', substr( $file_end, 0, 1 ) );
    my $file_end_sec = unpack( 'C', substr( $file_end, 1, 1 ) );
    $file_end_sec = ($file_end_sec << 8) + unpack( 'C', substr( $file_end, 2, 1 ) );
    $file_end_sec = ($file_end_sec << 8) + unpack( 'C', substr( $file_end, 3, 1 ) );
    $file_end_sec = ($file_end_sec << 8) + unpack( 'C', substr( $file_end, 4, 1 ) );
    $file_end_sec = (0xFFFFFFFF - $file_end_sec) + 1;

    if( $file_ident eq $ident && $file_end_1 == 1 && $file_end_sec == $file_sectors )
    {
      print "\"$file\": $file_sectors+1 data sectors\n";
      push @files, [$file, $file_sectors];
    } # if( $file_ident ...
    else
    {
      print "WARNING: Input file \"$file\" is not a valid ARCADEmini binary.\n";
    } # if( $file_ident ... else

    close InF;
  } # if( open( InF ...
  else
  {
    print "WARNING: Can't open input file \"$file\": $!\n";
  } # if( open( InF ... else ...
} # for( $file_no ...
print "\n";

# count data sectors

my $total_sectors = 0;
foreach (@files)
{
  my ($file, $file_sectors) = @{$_};
  my $start_sector = $total_sectors;
  $_ = [$file, $file_sectors, $start_sector];
  $total_sectors += $file_sectors + 1;
}
print "total: $total_sectors data sectors\n\n";

# write output file

my $out_file = $ARGV[@ARGV - 1];
open( OutF, "> " . $out_file ) || die "Can't open output file \"$out_file\": $!\n";
binmode OutF;
print OutF $ident;

my $sectors = 0;
my $header_sectors = int( (@files + 15) / 16 );
print "randomization header: $header_sectors data sectors\n\n";

# random number generation
my $mask = 0xFFFF;
$mask >>= 1 while( $mask >> 1 >= @files - 1 );
my $bytes = 12;
print OutF pack( 'CC', 0xD8, 0x00 ); # RND $0
print OutF pack( 'CCCC', 0x71, $mask & 0xFF, 0x00, 0x00 ); # AND mask_low, $0, $0
print OutF pack( 'CC', 0xD8, 0x01 ); # RND $1
print OutF pack( 'CCCC', 0x71, ($mask >> 8) & 0xFF, 0x01, 0x01 ); # AND mask_high, $1, $1

# jumps
my $j = 0;
while( $j < @files )
{
  for( $i = 0; $i < 16 and $j < @files; $i++, $j++ )
  {
    my ($file, $file_sectors, $start_sector) = @{@files[$j]};

    $bytes += 13;
    print OutF pack( 'CCCC', 0x72, 0x00, 0x01, 0x02 ); # OR $0, $1, $2
    print OutF pack( 'CCCC', 0x0E + ($bytes >> 8), $bytes & 0xFF, 0x02, 0x00 ); # JMP_NEQ @label, $2, 0
    my $jump_sectors = $start_sector + $header_sectors - $sectors;
    print OutF pack( 'C', 0x01 ); # JMP_SEC <destination>
    print OutF pack( 'C', ($jump_sectors >> 24) & 0xFF );
    print OutF pack( 'C', ($jump_sectors >> 16) & 0xFF );
    print OutF pack( 'C', ($jump_sectors >> 8) & 0xFF );
    print OutF pack( 'C', $jump_sectors & 0xFF );
    # @label:

    $bytes += 12;
    print OutF pack( 'CCCC', 0x61, 0xFF, 0x00, 0x00 ); # ADD -1, $0, $0
    print OutF pack( 'CCCC', 0x0E + ($bytes >> 8), $bytes & 0xFF, 0x00, 0xFF ); # JMP_NEQ @label2, $0, -1
    print OutF pack( 'CCCC', 0x61, 0xFF, 0x01, 0x01 ); # ADD -1, $1, $1
    # @label2:
  }

  if( $j >= @files )
  {
    my $jump_sectors = $sectors == 0 ? 0 : 0xFFFFFFFF - ($sectors - 1);
    $bytes += 6;
    print OutF pack( 'CC', 0x04, 0x0C ); # JMP_OFS_SEC sector 0 offset 12
    print OutF pack( 'C', ($jump_sectors >> 24) & 0xFF );
    print OutF pack( 'C', ($jump_sectors >> 16) & 0xFF );
    print OutF pack( 'C', ($jump_sectors >> 8) & 0xFF );
    print OutF pack( 'C', $jump_sectors & 0xFF );
  }
  for( ; $bytes < 512; $bytes++ ) { print OutF pack( 'C', 0x00 ); }
  $bytes = 0;
  $sectors++;
}

foreach (@files)
{
  my ($file, $file_sectors, $start_sector) = @{$_};

  open( InF, "< " . $file ) || die "WARNING: Can't open input file \"$file\": $!\n";
  binmode InF;

  seek( InF, 512, 0 );
  for( $i = 0; $i < $file_sectors; $i++ )
  {
    my $buffer;
    read( InF, $buffer, 512 );
    print OutF $buffer;
  }

  # jump to end
  my $jump_sectors = $total_sectors - ($start_sector + $file_sectors);
  print OutF pack( 'C', 0x01 );
  print OutF pack( 'C', ($jump_sectors >> 24) & 0xFF );
  print OutF pack( 'C', ($jump_sectors >> 16) & 0xFF );
  print OutF pack( 'C', ($jump_sectors >> 8) & 0xFF );
  print OutF pack( 'C', $jump_sectors & 0xFF );
  for( $i = 5; $i < 512; $i++ ) { print OutF pack( 'C', 0x00 ); }

  $sectors += $file_sectors + 1;

  close InF;
} # foreach (@files)

print "\"$out_file\": $sectors data sectors\n";

my $jump_sectors = 0xFFFFFFFF - ($sectors - 1);
print OutF pack( 'C', 0x01 );
print OutF pack( 'C', ($jump_sectors >> 24) & 0xFF );
print OutF pack( 'C', ($jump_sectors >> 16) & 0xFF );
print OutF pack( 'C', ($jump_sectors >> 8) & 0xFF );
print OutF pack( 'C', $jump_sectors & 0xFF );
for( $i = 5; $i < 512; $i++ ) { print OutF pack( 'C', 0x00 ); }

close OutF;

print "\n";
